"""
    
    Custom search content type
    
    Copyright 2006 Red Innovation
    
    A content type which displays custom search form
        - Searched fields can be edited on-the-fly
        - Form can be placed anywhere in the navigation tree, like normal content
        - Archetypes Edit backend can be used to customize form texts
        
    There is a special always present field "url" which points to the
    result object.
        
    See tutorial XXX how to add search indexes for the site, etc.        
    
"""

__author__  = 'Mikko Ohtamaa <mikko@redinnovation.com>'
__docformat__ = 'restructuredtext'

# Python imports
import logging  

# Zope 3rd party imports
from Products.AdvancedQuery import AdvancedQuery

# Plone imports
from Products.Archetypes.utils import capitalize
from Products.Archetypes import Field
from Products.Archetypes.public import *
from Products.ATContentTypes.content.base import ATCTContent
from Products.ATContentTypes.content.base import updateActions, updateAliases
from Products.ATContentTypes.content.schemata import ATContentTypeSchema
from Products.ATContentTypes.content.schemata import finalizeATCTSchema
from Products.ATContentTypes.content.base import ATCTOrderedFolder, ATCTFolder
from Products.ATContentTypes.content.folder import ATFolderSchema, ATFolder
from Products.ATContentTypes.content.base import updateActions, updateAliases
from Products.ATReferenceBrowserWidget.ATReferenceBrowserWidget import *
from Products.CMFCore.utils import getToolByName
from Products.validation import V_REQUIRED


# Plone 3rd party imports
from Products.DataGridField import DataGridField, DataGridWidget
from Products.DataGridField.Column import Column
from Products.DataGridField.SelectColumn import SelectColumn

# Local imports
from config import *
from permissions import *
from searchfields import *
from sorting import *
import vocabularies

# Our logger object
logger = logging.getLogger('CustomSearch')

schema=ATContentTypeSchema.copy() + Schema((
                                                    
        TextField('bodyText',               
               default="How to use this search for to locate foobar at our site...",
               widget=RichWidget(                   
                   label="Body text",
                   description="Search help text displayed to user",
                )               
           ),                                                                                                                    
                                            
        TextField('resultViewMacro',               
               default="custom_search_result_listing",
               widget=StringWidget(
                  label="The search result listing page template",
                  description="TAL page template macro which is used to render result listing",
               )
              
        ),        
                                            
        IntegerField('userSortKeyCount',               
               default="2",
               validators=(("isInt", V_REQUIRED),),               
               widget=IntegerWidget(
                  label="User available sort keys",
                  description="How many sort criteria the user can select for this search",
               )
              
        ),                                             
       
         DataGridField('fields',                       
                 widget = DataGridWidget(                  
                     description="Define fields which can be used in this search",
                     
                     columns ={
                             "name" : Column("Friendly name"),
                             "type" : SelectColumn("Search type", "getSearchTypes"),                             
                             "index" : SelectColumn("Catalog index", "getCatalogIndexesAndMetadata"), 
                             "default" : Column("Default value"), 
                             "visibility" : SelectColumn("Visibility", "getSearchFieldVisibilityTypes"), 
                             "mode" : SelectColumn("Mode", "getSearchFieldModes"), 
                             
                         }
                     ),
                 columns=(
                     'name',
                     'index',
                     'default',
                     'type',       
                     'visibility',
                     'mode'
                  ),  
           ),
                       
         DataGridField('sorting',                       
                 widget = DataGridWidget(                  
                     description="Define how sorting is done",
                     
                     columns ={
                             "name" : Column("Friendly name"),                                                          
                             "index" : SelectColumn("Catalog index", "getCatalogMetadatas"), 
                             "priority" : Column("Priority"),                              
                             "mode" : SelectColumn("Mode", "getSortKeyModes"), 
                             "order" : SelectColumn("Order", "getSortKeyOrders"),                                
                             
                         }
                     ),
                 columns=(
                     'name',
                     'index',
                     'priority',
                     'mode', 
                     'order'
                  ),                         
    )

))

finalizeATCTSchema(schema)

# Schema post-construction modifications
# Do changes necessary where our schema differs from standard ATCT schema
del schema["relatedItems"]
del schema["allowDiscussion"]


class CustomSearch(ATCTContent):
    """ Custom search 
    
    """
    
    # Special name for the internal search field storing the link to the object
    URL_SEARCH_FIELD = "url"
    
    portal_type = meta_type = 'CustomSearch' 
    archetype_name = typeDescription = 'Custom search'  
    content_icon   = 'custom_search_icon.gif'        
    schema = schema
        
    global_allow = True    
    
    security = ClassSecurityInfo()
        
    def getLayout(self):
        """ Get layout which is used for viewing this item
        
        Override BrowserMixin dynamic layout selecter        
        """
        return "custom_search_form_view"
    
    # Vocabulary accessors
            
    security.declarePublic("getSearchTypes")    
    def getSearchTypes(self):    
        return vocabularies.getSearchTypes()

    security.declarePublic("getSearchFieldVisibilityTypes")            
    def getSearchFieldVisibilityTypes(self):
        return vocabularies.getSearchFieldVisibilityTypes()   
        
    security.declarePublic("getSearchFieldModes")            
    def getSearchFieldModes(self):             
        """ Generate edit mode drop down menu """
        return vocabularies.getSearchFieldModes()
        
    security.declarePublic("getSortKeyModes")            
    def getSortKeyModes(self):             
        return vocabularies.getSortKeyModes()
        
    security.declarePublic("getSortKeyOrders")            
    def getSortKeyOrders(self):             
        return vocabularies.getSortKeyOrders()        
        

    security.declarePublic("getCatalogMetadata")    
    def getCatalogMetadatas(self):          
        """ Generate edit mode drop down menu 
        
        Returns portal_catalog indexes as DisplayList for SelectColumn.        
        """
        return self.getCatalogIndexesAndMetadata(indexes=True, metadatas=False)
        
    security.declarePublic("getCatalogIndexesAndMetadata")    
    def getCatalogIndexesAndMetadata(self, indexes=True, metadatas=True):      
        """ Generate edit mode drop down menu 
        
        Returns portal_catalog indexes as DisplayList for SelectColumn.
        
        - Contains indexes for search query
        - Contains metadata colums to get possible ImageSearchField image URL field        
        
        @param indexes True to List search indexes (for querying)
        @param metadats True to List metadatas (for result listing)
        """
        
        list = DisplayList()
        
        portal_catalog = getToolByName(self, 'portal_catalog') 
        
        # add indexes to the list        
        all = []
        
        if indexes:
            for idx in portal_catalog.indexes():
                all.append(idx)
        
        # add metadata to list
        # could be done with filter, map, etc, but 
        # i was lazy
        if metadatas:
            for meta in portal_catalog.schema():
                # prevent duplication
                if meta not in all:
                    all.append(meta)
        
        # Make DisplayList tuples from index/metadata entries
        for idx in all:
            list.add(idx, idx)
        
        
        return list
        
    security.declarePublic("getSortKeyManager")         
    def getSortKeyManager(self, REQUEST=None):
        """ TODO: Cache per-request SortKeyManagers """
        skm = SortKeyManager(self)        
        return skm
        
    security.declarePublic("getUserChoicesForSorting")         
    def getUserSortingVocabulary(self, REQUEST=None):
        return self.getSortKeyManager(REQUEST).getUserSortingVocabulary()
    
    security.declarePublic("getSearchFields")                  
    def getSearchFields(self, REQUEST=None):
        """ Convert DataGridField user inputted data to SearchField objects 
        
        @param REQUEST Cache parsed field data in REQUEST object. Not IMPLEMENTED.
        """
        
        fields = []
        
        rows = self.getFields()
        for r in rows:
            # fetch values from DataGridField row
            name = r["name"]
            index = r["index"]
            type = r["type"]
            default = r["default"]
            visibility = r["visibility"]
            mode = r["mode"] 
            
            # Set defaults for blanks
            if visibility == "":
                visibility = SearchField.VISIBILITY_QUERY_AND_RESULTS
                
            if mode == "":
                mode = SearchField.MODE_USER_EDITABLE
                
            if EXTENDED_INPUT_CHECKING:
                if index not in self.getCatalogIndexesAndMetadata():
                    raise AttributeError, "Tried to search against non-existing index %s" % index
                
            field = self.createSearchField(type, name, index, default, visibility, mode)
                        
            fields.append(field)   
            
        # Add internal URL field 
        # Otherwise we couldn't link in results
        fields.append( ValueSearchField(self, CustomSearch.URL_SEARCH_FIELD, "getURL", "", SearchField.VISIBILITY_NONE, SearchField.MODE_SPECIAL))

        return fields

    security.declareProtected(EXECUTE_CUSTOM_SEARCH_PERMISSION, "doCustomSearch")    
    def doCustomSearch(self, fields=None):
        """ Executes a custom search query 
        
        Builds AdvancedQuery from:
            - Indexes defined in the edit mode of CustomSearch
            - Values given in HTTP form displayed in the view mode of CustomSearch
            
        @return List of found items. Each item is dictionary of search field name : index value pairs
        """
        
        request = self.REQUEST
                    
        if fields == None:
            fields = self.getSearchFields()
            
        if len(fields) == 0:
            # TODO: Graceful error handling
            # display a page or something
            raise ValueError, "No search field defined"
        
        # Iterate through defined search fields.
        # Use SearchField object to read user given 
        # search values from the HTTP request form
        for field in fields:
            field.readRequest(request)
            
        skm = self.getSortKeyManager(request)
                                        
        query = self.constructSearchQuery(fields)
        if query == ():
            # TODO: Should not happen
            raise ValueError, "Empty search query"
        
        #print "query:" + str(query)
        portal_catalog = getToolByName(self, 'portal_catalog')
        
        sortSpec = skm.constructSortSpecification(request, self.getUserSortKeyCount())
        
        logger.debug("AdvancedQuery Search query spec:" + str(query))
        
        results = portal_catalog.evalAdvancedQuery(query, sortSpec)
        
        # format results to dictionary
        results = self.makeResultListing(results, request)
                
        logger.debug("Got results:" + str(results))
                        
        return results
        
        
    security.declareProtected(EXECUTE_CUSTOM_SEARCH_PERMISSION, "showResults")    
    def showResults (self):        
        """ Executes a custom search query and redirects user to the search results form 
        
        Search result form is rendered by custom_search_form_view
        when searched=true is set to HTTP GET parameters.
        """
                        
        request = self.REQUEST
        #print "url:" + url        
        #for key in request.keys():
        #    print "key:" + key + " val:" + str(request[key])
        
        # Redirect to self
        url = self.absolute_url()    
        
        # Forward HTTP GET parameters
        if request.has_key("QUERY_STRING"):
            url += "?" + request["QUERY_STRING"]
        
        # Add show results flag
        if not request.has_key("searched"):            
            url += "&searched=true"
                                       
        response = request.RESPONSE
        response.redirect(url)
        

    security.declarePrivate("makeResultListing")    
    def makeResultListing(self, results, request):        
        """ Convert ZCatalog brains objects to friendly list of dictionaries presentation
        
        @return sequence of dictionaries of search index value pairs
        
        """
        
        stp = getToolByName(self, 'portal_properties').site_properties
        view_action_types = stp.getProperty('typesUseViewActionInListings', ())
    
        listing = []
        
        # Go through list of ZCatalog brains (search index data snapshots)
        # and convert them to index : value dictionaries
        for brain in results:
            entry = {}
            for field in self.getSearchFields():     
                                                
                # should this field be included in search results
                if field.isVisibleInResults() or field.getMode() == SearchField.MODE_SPECIAL:  
                    
                    # if user hasn't entered friendly name we get troubles...
                    if field.getFriendlyName() != None and field.getFriendlyName() != "":
                        
                        # TODO: Make class URLSearchField to handle this case
                        if(field.getFriendlyName() == CustomSearch.URL_SEARCH_FIELD):
                            # Item URL needs some magic (taken from PloneTool)
                            # Add /view to URL if needed
                            item_url = (brain.portal_type in view_action_types and
                                    brain.getURL() + '/view') or brain.getURL()
                      
                            entry[CustomSearch.URL_SEARCH_FIELD] = item_url
                        else:
                            # Convert brain value to page template
                            # suitable form
                            field.fetchValueBromBrain(entry, brain, request)
            
            # proactive bug preveting - don't let bad things fall in
            assert CustomSearch.URL_SEARCH_FIELD in entry
            
            listing.append(entry)
            
        return listing
        

    security.declarePrivate("constructSearchQuery")            
    def constructSearchQuery(self, fields):
        """ Craft AdvancedQuery search for portal_catalog """
        
        subqueries = []
                        
        # Use SearchField method to get AdvancedQuery object    
        for field in fields:
            query = field.createQuery()
            if query != None:
                subqueries.append(query)
                
        # Concatenate all conditions with final AND
        return AdvancedQuery.And(*subqueries)
    
        
    security.declarePrivate("createSearchField")
    def createSearchField(self, type, name, index, default, visibility, mode):
        """ Constructs a search field object from tag 
        
        A factory for search fields. Override to add your
        own search field types.
        """
                
        # Construct SearchField object
        if type == "value":
            field = ValueSearchField(self, name, index, default, visibility, mode)
        elif type == "minmax":
            field = MinMaxSearchField(self, name, index, default, visibility, mode)
        elif type == "list":
            field = ListSearchField(self, name, index, default, visibility, mode)                
        elif type == "image":
            field = ImageSearchField(self, name, index, default, visibility, mode)                                
        elif type == "path":
            field = PathSearchField(self, name, index, default, visibility, mode)                                            
        elif type == "boolean":
            field = BooleanSearchField(self, name, index, default, visibility, mode)
        elif type == "":
            raise AttributeError, "Missing field definition at object " + self.title_or_id()
        else:
            raise AttributeError, "Unknown search field definition " + type + " at object " + self.title_or_id()
            
        return field
                
        
    
registerType(CustomSearch, PROJECTNAME)

