"""
    
    
    Different search fields
   
    Copyright 2006 Red Innovation
    
"""

__author__  = 'Mikko Ohtamaa <mikko@redinnovation.com>'
__docformat__ = 'restructuredtext'

# Python imports
import urllib
import types
import logging

# Zope imports
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from Globals import InitializeClass

try:
    from zope.component import getMultiAdapter
except:
    # Plone 2.1.x series don't have getMultiAdapter
    # this is required for PathSearchField result list generating
    # not used anywhere else
    getMultiAdapter = None

# Zope 3rd party imports
from Products.AdvancedQuery import AdvancedQuery

# Plone imports
from Products.Archetypes.public import *
from Products.CMFCore.utils import getToolByName

# Our logger object
logger = logging.getLogger('CustomSearch')
       
class SearchField:
    """ CustomSearch search field definition in easy-to-manipulate form 
    
    Base class for all search field declarations.
    
    Life cycle: one HTTP request. Note that data inside SearchField
    object varies during the different stage of the processing the HTTP
    request. This is a bad practice from modelling view point, but 
    makes things simple (we need less objects).
    
    The life span of SearchField
        - Created when form is read
            * Persisntent CustomSearch data is combined with read form values
        - Deleted when HTTP response is closed    
        
    """
    
    # Tag prefixing all page fields indicating they should be searched
    MAGIC_SEARCH_FIELD_TAG = "searchfield_"
    
    # User can enter this search value
    MODE_USER_EDITABLE = "user_editable"
    
    # Always use search value entered by manager
    MODE_FIXED_VALUE = "fixed_value"
    
    # For internal use of CustomSearch
    MODE_SPECIAL = "special"

    # Don't display this field as fillable or its value in results    
    VISIBILITY_NONE = "none"    
    
    VISIBILITY_RESULTS = "result"
    VISIBILITY_QUERY = "query"
    VISIBILITY_QUERY_AND_RESULTS = "both"
    
    # query macro arg in __init__ keywords
    QUERY_MACRO_ARG = "queryMacro"
    
    # result macro arg in __init__ keywords
    RESULT_MACRO_ARG = "resultMacro"
    
    security = ClassSecurityInfo()
    
    def __init__(self, context, friendlyName, index, default, visibility, mode, **kwargs):  
        """ Initialize search field
        
        @param context CustomSearch object creating this query
        @param friendlyName User visible name
        @param index portal_catalog index name this field searches
        @param default HTML form default value for this field
        @param visibility Is this field shown in the search query, results or both
        @param mode Can user enter values for this search field or does it use fixed value 
        
        @param queryMacro optional, macro used to render query field
        @param resultMacro optional, macro used to render data in result listing
        
        """
                
                
        self.context = context
        self.friendlyName = friendlyName
        self.index = index        
        self.default = default
        self.visibility = visibility
        self.mode = mode
        
        # self.value = parsed search input
        # by default, there is no input for this search field
        # until we have succesfully read somethign from form
        # or default
        self.value = None
        
        if SearchField.QUERY_MACRO_ARG in kwargs:
            self.queryMacro = kwargs[SearchField.QUERY_MACRO_ARG]
        else: 
            self.queryMacro = "custom_search_field"
                    
        if SearchField.RESULT_MACRO_ARG in kwargs:
            elf.resultMacro = kwargs[SearchField.RESULT_MACRO_ARG]
        else:
            self.resultMacro = "custom_search_field"

        # proactive bug preveting - don't let bad types fall in
        assert type(self.friendlyName) in types.StringTypes
        assert type(self.index) in types.StringTypes
        assert type(self.default) in types.StringTypes
        assert type(self.visibility) in types.StringTypes
        assert type(self.mode) in types.StringTypes
        assert type(self.queryMacro) in types.StringTypes
        assert type(self.resultMacro) in types.StringTypes
          
    
    def getResultMacro(self):
        return "custom_search_field"
        
    security.declarePublic("getFriendlyName")
    def getFriendlyName(self):
        """ User visible name in HTML code """
        return self.friendlyName

    security.declarePublic("getSearchIndexId")
    def getSearchIndexId(self):
        """ Which portal_catalog index is used to match this field values """
        return self.index    
    
    security.declarePublic("getDefaultValue")
    def getDefaultValue(self):
        """ Use this value in HTML forms when user enter the form first time """
        return self.default
    
    security.declarePublic("getFormFieldName")    
    def getFormFieldName(self):
        """ <input name=> used in HTML forms """
        return SearchField.MAGIC_SEARCH_FIELD_TAG + self.index
    
    security.declarePublic("getMode")    
    def getMode(self):
        """ Return "user_edit" or "fixed_value" """
        return self.mode
    
    security.declarePublic("getVisibility")    
    def getVisibility(self):
        """ Return one of VISIBILITY_xxx """
        return self.visibility
        
    security.declarePublic("getQueryMacro")
    def getQueryMacro(self):
        """ TAL template used to render input fields 
        
        See custom_search_form_body.pt
        """
        return self.queryMacro

    security.declarePublic("getResultMacro")
    def getResultMacro(self):
        """ TAL template used to render input fields 
        
        See custom_search_result_listing.pt
        """
        return self.resultMacro
        
    security.declarePublic("readRequest")
    def readRequest(self, request):
        """ Read submitted form and get self.value
        
        If search field mode doesn't allow user editing
        for this value, ignore form and use field's default value.        
        """
        raise AttributeError, "Use subclass"
    
    security.declarePublic("isVisibleInResults")        
    def isVisibleInResults(self):
        """ Return true if field should be shown in results """
        return self.visibility == self.VISIBILITY_RESULTS or self.visibility == self.VISIBILITY_QUERY_AND_RESULTS

    security.declarePublic("isVisibleInQuery")        
    def isVisibleInQuery(self):
        """ Return true if field should be shown on search query page """
        return (self.getMode() == SearchField.MODE_USER_EDITABLE) and (self.visibility == self.VISIBILITY_QUERY or self.visibility == self.VISIBILITY_QUERY_AND_RESULTS)
        
    security.declarePublic("createQuery")
    def createQuery(self):
        """ Get AdvancedQuery parameters for this field """
        raise AttributeError, "Use subclass"
    
    def fetchValueBromBrain(self, entry, brain, request):
        """ Read value from returned search index snapshot
        
        Place value into a dictionary in a format
        page templates can understand.
        
        @param entry Dictionary which stores search hit entry data going to page templates
        @param brain ZCatalog brain which was returned from search query
        @param request HTTPRequest
        """
        
        logger.debug("Fetching value for " + self.getFriendlyName() + " from index " + self.getSearchIndexId() )
        
        if hasattr(brain, self.getSearchIndexId()):
            # brain contains indexes field we are looking for
            entry[self.getFriendlyName()] = getattr(brain, self.getSearchIndexId())
        else:
            # this brain lacks indexed field we are looking for
            # return empty string                                
            entry[self.getFriendlyName()] = ""        
    
# Initializes class security
InitializeClass(SearchField)      
    
class ValueSearchField(SearchField):
    """ Search for exact value match """
          
    def readRequest(self, request, fieldName = None):        
        """ Read search parameters from HTTP request
        
        @param request HTTPRequest object
        @param fieldName override default field name generation logic, 
               for example for MinMax field which has two input fields
        """
        
        # force in site manager's value
        # when should not be able to modify the field
        if self.getMode() == SearchField.MODE_FIXED_VALUE:
            self.value = self.default
            return self.value
        
        if fieldName == None:
            fieldName = self.getFormFieldName()        
    
        
        # HTTP request has input field for us?
        if request.has_key(fieldName):
            # read input, strip extra spaces
            self.value = request[fieldName].strip()
            return self.value
        
        # HTTP request did not have input for us
        return None

    def createQuery(self): 
        if self.value != None and self.value != "":
            return AdvancedQuery.Eq(self.index, self.value)
        return None
    
# Initializes class security
InitializeClass(ValueSearchField)    

class ListSearchField(ValueSearchField):
    """ Value must be one of listed 
    
    Comma separate list of possible values
    """
    
    def createQuery(self):
        """ Splits comma separated list to sequence, trims values for extra spaces """
                
        if self.value == None or self.value == "":
            return None            
        
        values = self.value.split(",")
        values = map(lambda x: x.strip(), values)
                        
        return AdvancedQuery.In(self.index, values)
   
# Initializes class security
InitializeClass(ListSearchField)   


class PathSearchField(ValueSearchField):
    """ Match searches against portal path index
    
    Slash separated list for the site path, starting from the
    site root.
    
    E.g. /customer-service/knowledge-base
    
    For details about possible parameters, see ExtenedPathIndex
        
    """
    
    def __init__(self, context, friendlyName, index, default, visibility, mode, **kwargs):
                           
         ValueSearchField.__init__(self, context, friendlyName, index, default, visibility, mode, **kwargs)
         
         if SearchField.RESULT_MACRO_ARG in kwargs:
             # User has overriden macro 
             self.resultMacro = kwargs[SearchField.RESULT_MACRO_ARG]
         else:
             # User default macro
             self.resultMacro = "custom_search_path_field"        
    
    def createQuery(self):
        """ Splits comma separated list to sequence, trims values for extra spaces """
                
        # ignore empty search input
        if self.value == None or self.value == "":
            return None            
                                
        portal_url = getToolByName(self.context, "portal_url")
        
        # add required /
        if self.value.startswith("/"):
            target = self.value
        else:
            target = "/" + self.value

        # Since ExtendedPathIndex operates in Zope context
        # (it can cross multi site boundaries inside one Zope) 
        # we need to be sure that we start searching
        # using our portal root as a base
        
        return AdvancedQuery.Generic(self.index, portal_url.getPortalPath() + target)
        
    def fetchValueBromBrain(self, entry, brain, request):
        """ Read value from returned search index snapshot
        
        Use Plone utils to get breadcrumbs for the object.
        This can be slow since index are not used for the operation,
        but breadcrumbs are generated using acquiring chain per item.
        """        
        
        if getMultiAdapter != None:
            
            # Plone 2.5
        
            context = brain.getObject()
                    
            view = getMultiAdapter((context, request), name='breadcrumbs_view')
            crumbs = view.breadcrumbs()
            
            logger.debug("Got crumbs:" + str(crumbs))
            entry[self.getFriendlyName()] = crumbs
        
        else:
            logger.error("Plone 2.5 is needed for PathSearchField result listing")
            entry[self.getFriendlyName()] = []

        
        
        
   
# Initializes class security
InitializeClass(PathSearchField)        
    

class BooleanSearchField(ValueSearchField):
    """ Match values against boolean objects in index 
    
    """
    
    def __init__(self, context, friendlyName, index, default, visibility, mode, **kwargs):
                           
         ValueSearchField.__init__(self, context, friendlyName, index, default, visibility, mode, **kwargs)
         
         if SearchField.QUERY_MACRO_ARG in kwargs:
             # User has overriden macro 
             self.queryMacro = kwargs[SearchField.QUERY_MACRO_ARG]
         else:
             # User default macro
             self.queryMacro = "custom_search_boolean_field"    

    def createQuery(self): 
        if self.value != None and self.value != "":
            
            target = False
            self.value.strip()
            
            logger.debug("Boolean search field " + self.friendlyName + " entry value was " + self.value)

            # Did user put choice in the field?
            if self.value != "":
            
                if self.value in [ "true", "True", "1", "on", "yes"]:
                    target = True
                
                return AdvancedQuery.Eq(self.index, target)
        
        return None
    
# Initializes class security
InitializeClass(BooleanSearchField)       

class MinMaxSearchField(ValueSearchField):
    """ An index value must be within certain limits
    
    Suitable for numbers only.
    
    Either min limit or max limit can be empty, thus being ignored.    
    
    For default value, min and max are given as a comma separated list
    """
    
    MIN_TAG = "min_"    
    MAX_TAG = "max_"
    
    security = ClassSecurityInfo()    
          
    def __init__(self, context, friendlyName, index, default, visibility, mode, **kwargs):
        
        # set our default query macro
        if SearchField.QUERY_MACRO_ARG not in kwargs:
            kwargs[SearchField.QUERY_MACRO_ARG] = "custom_search_minmax_field"
        
        ValueSearchField.__init__(self, context, friendlyName, index, default, visibility, mode, **kwargs)
        
        # Read default min, max as comma separated list
        values = default.split(",")
        values = map(lambda x: x.strip(), values)        
        
        if len(values) > 0:
            self.minDefault = values[0]
        else:
            self.minDefault = 0
            
        if len(values) > 1:
            self.maxDefault = values[1]
        else:
            self.maxDefault = 9999999
        
    def getMacro(self):
        return "custom_search_minmax_field"
    
    security.declarePublic("getFormFieldNameMin")
    def getFormFieldNameMin(self):
        return MinMaxSearchField.MIN_TAG + self.index

    security.declarePublic("getFormFieldNameMax")
    def getFormFieldNameMax(self):
        return MinMaxSearchField.MAX_TAG + self.index
    
    def readRequest(self, request):
         """ Read in HTML form fields """
        
         if self.getMode() == SearchField.MODE_FIXED_VALUE:
             self.min = self.minDefault
             self.max = self.maxDefault
         else:        
            self.min = ValueSearchField.readRequest(self, request, self.getFormFieldNameMin()) 
            self.max = ValueSearchField.readRequest(self, request, self.getFormFieldNameMax())    
                
    def createQuery(self): 
        
        minq = None
        maxq = None
               
        # which bounds we need to set
        if self.min != None and self.min != "":            
            try:
                minq = float(self.min)
            except:
                # TODO fails silently on a bad number
                pass
            
        if self.max != None and self.max != "":        
            try:
                maxq = float(self.max)
            except:
                # TODO fails silently on a bad number                
                pass
            
        #print "query %s - %s" % (self.min, self.max)
                
        # Check do we have  min bound, max bound or both            
        if minq != None and maxq != None:
            #print "woo:" + str(minq) + " " + str(maxq)
            return AdvancedQuery.Between(self.index, minq, maxq)
        elif minq != None:
            return AdvancedQuery.Ge(self.index, minq)
        elif maxq != None:
            return AdvancedQuery.Le(self.index, maxq)
        else:
            return None
    
# Initializes class security
InitializeClass(MinMaxSearchField)

class ImageSearchField(ValueSearchField):
    """ Display image in search results
    
    - Image field doesn't contain query parameters, it's always
      hidden on the query form
      
    How to use:
    
        1. The image for the result is fetched using portal_catalog metadata. 
           Search field parameter 'index' is metadata column name.
                
            * Create a metadata column 'getPrimaryThumbnail'
        
        2. Create a method MyContentType.getPrimaryThumbnail(). This method returns to URL for the image. 
           It could be something like this for folderish object (returns the first image in the folder)::
    
                def getPrimaryThumbnail(self):
                    ''' Retrieves the HTML tag for thumbnail image used in search result listings 
                                    
                    Return the first image in the content listing of folderish object,
                    so this method applies for folderish objects only. If no images are 
                    available, return None.
                    
                    This function might be performance expensive, it calls actcListAlbum
                    Python script.
                    
                    You need to have getPrimaryThumbnail in portal_catalog metadata.
                    '''
                    result = self.atctListAlbum(images=True, folders=False, subimages=False, others=False)
                    
                    if "images" in result:
                        images = result["images"]
                        if len(images) > 0:
                            tag = images[0].tag(scale='thumb', title=images[0].Description())
                            print "tag is:" + tag
                            return tag                
                        else:
                            return None
 
    
    """
    
    def __init__(self, context, friendlyName, index, default, visibility, mode, **kwargs):
         
         # ignore user set visibility and mode
         visibility = SearchField.VISIBILITY_RESULTS
         mode = SearchField.MODE_SPECIAL
                      
         ValueSearchField.__init__(self, context, friendlyName, index, default, visibility, mode, **kwargs)
         
         if SearchField.RESULT_MACRO_ARG in kwargs:
            self.resultMacro = kwargs[SearchField.RESULT_MACRO_ARG]
         else:
            self.resultMacro = "custom_search_image_field"

                 
# Initializes class security
InitializeClass(ImageSearchField)       


    
    