"""
    
    
    Sort key definition
   
    Copyright 2006 Red Innovation
    
"""

__author__  = 'Mikko Ohtamaa <mikko@redinnovation.com>'
__docformat__ = 'restructuredtext'

# Python imports
import urllib
import types
import logging

# Zope imports
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from Globals import InitializeClass

# Zope 3rd party imports
from Products.AdvancedQuery import AdvancedQuery

# Plone imports
from Products.Archetypes.public import *
from Products.CMFCore.utils import getToolByName

from config import *

# Our logger object
logger = logging.getLogger('CustomSearch')
       
class SortKey:
    """ Define sort key 
    
    Life cycle: one HTTP request. Note that data inside SortKey
    object varies during the different stage of the processing the HTTP
    request. This is a bad practice from modelling view point, but 
    makes things simple (we need less objects).
    """ 

    # Which priority is given for user selected sort keys
    USER_SORT_KEY_PRIORITY_BASE = 100    
        
    # Tag prefixing all page fields indicating they should be searched
    # add 0...n to this tag to get HTTP field name
    HTML_SORT_KEY_FIELD_TAG = "sortkey_"
    
    # User can enter this search value
    MODE_AVAILABLE_FOR_USER = "user_choosable"
    
    # Always use search value entered by manager
    MODE_FORCED = "forced_in"
    
    # For internal use of CustomSearch
    MODE_SPECIAL = "special"
        
    ORDER_ASCENDING = "ascending"
    
    ORDER_DESCENDING= "descending"
    
    security = ClassSecurityInfo()
    
    def __init__(self, context, index, friendlyName, priority, mode, order, **kwargs):  
        """ Initialize search field
        
        @param context CustomSearch object creating this query
        @param index portal_catalog index used
        @param friendlyName User visible name        
        @param mode one of MODE_XXX
        @param priority How high we rank this index (0 == highest)            
        """
                                
        self.context = context
        self.friendlyName = friendlyName
        self.index = index        
        self.mode = mode
        self.order = order
        
        try:
            self.priority = int(priority.strip())
        except:
            # TODO: Fix silent fail
            self.priority = 9999
        
        # proactive bug preveting - don't let bad types fall in
        assert type(self.friendlyName) in types.StringTypes
        assert type(self.index) in types.StringTypes
        assert type(self.priority) == types.IntType
        assert type(self.mode) in types.StringTypes        
        assert type(self.order) in types.StringTypes  

                  
    security.declarePublic("getFriendlyName")
    def getFriendlyName(self):
        """ User visible name """
        return self.friendlyName

    security.declarePublic("getSearchIndexId")
    def getIndex(self):
        """ Which portal_catalog index is used to match this field values """
        return self.index    
        
    security.declarePublic("getMode")    
    def getMode(self):
        return self.mode
    
    security.declarePublic("getOrder")    
    def getOrder(self):
        return self.order
    
    security.declarePublic("getPriority")    
    def getPriority(self):
        return self.priority
    
    security.declarePublic("getPriority")    
    def overridePriority(self, userPriority):
        """ Used to force user selected keys after forced sort key indexes """
        self.priority = userPriority
    
    def isAvailableForUser(self):    
        """ True if this key should appear in the end user sort selection """
        return self.mode == SortKey.MODE_AVAILABLE_FOR_USER
    
    def createSortSpec(self):
        """ Create advanced query sort specification object
        
        It's tuple (indexname, "asc" or "desc")
        """
        return (self.getIndex(), self.getOrder() == SortKey.ORDER_ASCENDING and "asc" or "desc")

# Initializes class security
InitializeClass(SortKey)   
                                                  
class SortKeyManager:                 
    """ Manage sort key input and construct sort specification for search query
    
    - Convert persistent DataGridField data to SortKey objects
    
    - Read sort key data from HTTP form request and combine it with 
    information provided by CustomSearch settings.

    """
    
    security = ClassSecurityInfo()    
    
    def __init__(self, context):
        """ Create sort key manager. """
        # Tranform persistent edit data from DGF grid to sort key objects
        self.sortkeys = self.initializeSortKeys(context)
    
    security.declarePublic("getSortKeys") 
    def getSortKeys(self):
        return self.sortkeys
                  
    def initializeSortKeys(self, context):
        """ Convert DataGridField persistent data to temporary SortKey objects 
        
        
        @param context CustomSearch object
        """
        
        keys = []
        
        rows = context.getSorting()
        for r in rows:
            # fetch values from DataGridField row
            name = r["name"]
            index = r["index"]
            mode = r["mode"] 
            priority = r["priority"]            
            order = r["order"]               
                        
            # DGF selection list box empty selection
            if mode == "":
                mode = SortKey.MODE_AVAILABLE_FOR_USER
                
            if order == "":
                order = SortKey.ORDER_ASCENDING
                
            if EXTENDED_INPUT_CHECKING:
                if index not in context.getCatalogIndexesAndMetadata():
                    raise AttributeError, "Tried to search against non-existing index %s" % index
                
            key = SortKey(self, index, name, priority, mode, order)
                        
            keys.append(key)   
            
        return keys
   
    security.declarePublic("getUserChoicesForSorting")         
    def getUserSortingVocabulary(self):
        """ Return selection list vocabulary for sort keys an user can choose """
        keys = self.getSortKeys()
        
        userKeys = DisplayList()
        
        for k in keys:
            if k.isAvailableForUser():
                # Since the same index can be used in two different 
                # sort keys we need to use friendly name as a key
                userKeys.add(k.getFriendlyName(), k.getFriendlyName())
                
        return userKeys    
      
    security.declarePublic("countSortKeysAvailableForUsers")      
    def countSortKeysAvailableForUsers(self):                                        
        """ Define how many max. sort keys the custom search object allows """
        n = 0
        for sk in self.sortkeys:
            if sk.isAvailableForUser():
                n += 1
                
        return n
    
    def lookUpSortKey(self, friendlyName):
        """ Return SortKey object assigned to a friendly name """
        for sk in self.sortkeys:
            if sk.getFriendlyName() == friendlyName:
                return sk
        return None
                        
    def readUserGivenSortKeys(self, request, maxUserKeys):
        """ Read search form data to see if this is one of selected sort key(s) and which priority it has 
        
        Return SortKey objects, in the order 
        """
               
        userkeys = []
                
        for n in range(0, maxUserKeys):
            sortKeyName = SortKey.HTML_SORT_KEY_FIELD_TAG + str(n)
            if request.has_key(sortKeyName):
                
                choice = request.get(sortKeyName)                
                key = self.lookUpSortKey(choice)
                
                if key == None:
                    raise ValueError, "Custom Search object doesn't define a sort key for name " + sortKeyName

                userkeys.append(key)
                
        return userkeys                
                
    def constructSortSpecification(self, request, maxUserKeys):
        """ Craft a sort specification for the request """
        
        # reset all priorities for user selectable sort keys to infinite
        for sk in self.sortkeys:
            if sk.getMode() == SortKey.MODE_AVAILABLE_FOR_USER:
                sk.overridePriority(9999)
                            
        # priorize user inputted keys 
        # this overrides infinity set above if the key is chosen
        if request != None:
             
            userkeys = self.readUserGivenSortKeys(request, maxUserKeys)                   
         
            n = SortKey.USER_SORT_KEY_PRIORITY_BASE
            for key in userkeys:
                key.overridePriority(n)
                n += 1
                
        # put sort keys to order based on the sort key priority
        self.sortkeys.sort(cmp=lambda x,y: x.getPriority() - y.getPriority())                        
        
        # Create AdvancedQuery sort spec list    
        spec = []
        for key in self.sortkeys:
            # don't allow duplicates 
            
            newSpec = key.createSortSpec()
            
            # Check that if the index already
            # exists in the spec             
            alreadyThere = False    
            for defined in spec:
                # Some spec item prior this already
                # has reseved this index
                if defined[0] == newSpec[0]:
                    alreadyThere = True
                    break
            
            if not alreadyThere:
                spec.append(newSpec) 
                
        logging.debug("Created sort specification " + str(spec))
        
        return spec
   
# Initializes class security
InitializeClass(SortKeyManager)   