"""
    
    Unit testing for Custom Search product
    
    Copyright 2006 Red Innovation
    
    Depends on PloneInstall product
    
"""

__author__  = 'Mikko Ohtamaa <mikko@redinnovation.com>'
__docformat__ = 'restructuredtext'

# Python imports
import inspect
import os, sys
import types
import logging

# Output debug log
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))    

# Zope imports
from AccessControl import getSecurityManager, ClassSecurityInfo, Unauthorized
from AccessControl.Permission import Permission
from Testing import ZopeTestCase
from Testing.ZopeTestCase.ZopeLite import *

# Set verbose security policy so we have more detailed information
# about security violations
from AccessControl.SecurityManager import setSecurityPolicy
from AccessControl.ImplPython import ZopeSecurityPolicy
setSecurityPolicy(ZopeSecurityPolicy(verbose=True))

# Plone imports
from Products.ATContentTypes.content.image import ATImage
from Products.Archetypes.tests.utils import Dummy
from Products.Archetypes.tests.utils import mkDummyInContext
from Products.Archetypes.public import DisplayList
from Products.CMFPlone.tests import PloneTestCase
from Products.CMFPlone.tests.PloneTestCase import PloneTestCase
from Products.CMFCore.utils import getToolByName

# Plone 3rd party imports
from Products.PloneInstallation.public import InstallationContext, CatalogInstaller, InstallationRunner

# Local imports
from Products.customsearch.config import *
from Products.customsearch.permissions import *
from Products.customsearch.searchfields import *
from Products.customsearch.sorting import *
  
ZopeTestCase.installProduct("DataGridField")
ZopeTestCase.installProduct("AdvancedQuery")
ZopeTestCase.installProduct("customsearch")

# Portal base

# Test content classes used

class DummyContent(ATImage):
    pass

class PriceDummyContent(ATImage):
    """ Test content type  which has two different indexable (FieldIndex) attributes """
    
    security = ClassSecurityInfo()
    
    security.declarePublic("setPrice")
    def setPrice(self, price):
        self.price = price
    
    security.declarePublic("getPrice")
    def getPrice(self):
        return self.price
    
    security.declarePublic("setAmount")
    def setAmount(self, amount):
        self.amount = amount
    
    security.declarePublic("getAmount")
    def getAmount(self):
        return self.amount    
    
class ImageSearchTestContent(ATImage):
    
    security = ClassSecurityInfo()
    
    MAGIC = "blaa"
    
    security.declarePublic("getThumbnail")
    def getThumbnail(self, **kwargs):
        print "called:" + str(kwargs)
        return self.MAGIC
        
class TestCustomSearch(PloneTestCase):
    """ Unit testing cases for real-estate management
       
    - Test that product can be succesfully installed 
    """
        
    def afterSetUp(self):
        PloneTestCase.afterSetUp(self)
        # quick install required products       
                
        self.loginOwner()       
        
        qi = self.portal.portal_quickinstaller
        qi.installProduct('DataGridField')
        qi.installProduct('customsearch')
        
        
        
    def testCustomSearch(self):   
        
        self.loginOwner()
        
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },
          {
          "type" : "list",
          "name" : "sampleList",
          "index" : "Title",
          "default" : "sample, sample2",
          },                  
          {
          "type" : "boolean",
          "name" : "Boolean",
          "index" : "is_folderish",
          "default" : "true",
          },                      
        ]

        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        self.portal.sampleSearch.setFields(fields)      
        self.portal.sampleSearch.setTitle("Search property")                  
        self.portal.sampleSearch.reindexObject()       
                
        
        self.portal.sampleSearch.doCustomSearch()
        
        pass    
    
    def testGetCatalogIndexes(self):   
        """ Test returning catalog indexes as a display list """
        self.loginOwner()
        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")  
        results = self.portal.sampleSearch.getCatalogMetadatas() 
        results = self.portal.sampleSearch.getCatalogIndexesAndMetadata() 
                
                
        # convert DisplayList to string representation
        # and seek for the key string
        self.failUnless("('Title', 'Title')" in str(results))
        
    def testCreateMinMaxField(self):
        """ See that MinMaxSearchField get/set things work """

        name = "test"
        index = "Title"
        default = ""
        visibility = SearchField.VISIBILITY_QUERY_AND_RESULTS
        mode = SearchField.MODE_USER_EDITABLE
        minmax = MinMaxSearchField(self.portal, name, index, default, visibility, mode)
        
        self.failUnless(minmax.getQueryMacro() == "custom_search_minmax_field")
        self.failUnless(minmax.getResultMacro() == "custom_search_field")
        
        minmax = MinMaxSearchField(self.portal, name, index, default, visibility, mode, queryMacro="xxx")
        self.failUnless(minmax.getQueryMacro() == "xxx")        
        
    
    def testMinMaxSearch(self):   
        """ Test min max search field 
        
        See that range limitation works propeply
        """
    
        self.loginOwner()
        
        self.createTestCatalogsAndObjects()
        
        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
                
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },                  
          {
          "type" : "minmax",
          "name" : "minmax",
          "index" : "getPrice",
          "default" : "0, 1000",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },
        ]

        self.portal.sampleSearch.setFields(fields)      
  
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()
            
        self.failUnless(len(results) == 1, "got results:" + str(results))
        
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },
          {
          "type" : "minmax",
          "name" : "minmax",
          "index" : "getPrice",
          "default" : "0, 3000",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },
        ]
        
        # Now execute custom search
        self.portal.sampleSearch.setFields(fields)            
        results = self.portal.sampleSearch.doCustomSearch()
        
        self.failUnless(len(results) == 2, "got results:" + str(results))
        
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },
          {
          "type" : "minmax",
          "name" : "minmax",
          "index" : "getPrice",
          "default" : "5000, 6000",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },
        ]

        # Now execute custom search
        self.portal.sampleSearch.setFields(fields)            
        results = self.portal.sampleSearch.doCustomSearch()        
        self.failUnless(len(results) == 0, "got results:" + str(results))
        

    
    def testShowResults(self):   
        """ Test min max search field """
    
        self.loginOwner()
        
        fields = [
          {
          "type" : "minmax",
          "name" : "minmax",
          "index" : "Title",
          },
        ]

        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        self.portal.sampleSearch.setFields(fields)      
        
        self.portal.sampleSearch.showResults()                
        pass       
    
    def testSearchField(self):
        """ Create value based search field """
        
        self.loginOwner()
        
        fields = [
          {
          "type" : "value",
          "name" : "test",
          "index" : "Title",
          },
        ]

        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        self.portal.sampleSearch.setFields(fields)     
        
    def testListingSeachField(self):
        """ Test that we can match by multiple value in list search """
                   
        self.loginOwner()
        
        fields = [
          # test query to return content types that are comma listed
          # should return URLs http://nohost/portal/news and http://nohost/portal/dummy
          {
          "type" : "list",
          "name" : "Content type",
          "index" : "Type",
          "default" : "ImageSearchTestContent, DummyContent",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },          
        ]
        
        # create a custom search
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        self.portal.sampleSearch.setFields(fields)   
                       
        # Now create the test object
        mkDummyInContext(ImageSearchTestContent, oid='dummy1', context=self.portal,
                                       schema=ATImage.schema)
        
        mkDummyInContext(DummyContent, oid='dummy2', context=self.portal,
                                       schema=ATImage.schema)
        
        # Make sure object is catalog
        self.portal.dummy1.reindexObject()
        self.portal.dummy2.reindexObject()
        
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()
        
        #print "list test results:" + str(results)
                            
        self.failUnless(len(results) == 2, "got %d results " % len(results))
        self.failUnless(self.isInResults(results, "/dummy1"), "got results:" + str(results))
        self.failUnless(self.isInResults(results, "/dummy2"), "got results:" + str(results))
                
    def testImageSeachField(self):
        """ Test that cataloged image tag is returned propeply """                   
        self.loginOwner()
        
        fields = [
          {
          "type" : "image",
          "name" : "image",
          "index" : "getThumbnail",
          },
          # force to return Dummy types only
          {
          "type" : "list",
          "name" : "Content type",
          "index" : "Type",
          "default" : "ImageSearchTestContent",
          "mode" : SearchField.MODE_FIXED_VALUE,
          "visibility" : SearchField.VISIBILITY_RESULTS,
          },          
        ]
        
        # create a custom search
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        self.portal.sampleSearch.setFields(fields)   
                       
        # create a custom portal_catalog search index
        indexes = [
               ("getThumbnail", "FieldIndex", None),                      
               ]
        
        metadata = ("getThumbnail",)
        
        runner = InstallationRunner(CatalogInstaller([], metadata),)
        
        ic = InstallationContext(self.portal, globals())
  
        # Runs the installation ans return the report
        runner.install(ic, auto_reorder=1)    
        
        # Now create the test object
        mkDummyInContext(ImageSearchTestContent, oid='dummy', context=self.portal,
                                       schema=ATImage.schema)
        
        # Make sure object is catalog
        self.portal.dummy.reindexObject()
        
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()
        
        #print "image results:" + str(results)
               
        self.failUnless(len(results) == 1, "got %d results " % len(results))
        self.failUnless(results[0]["image"] == ImageSearchTestContent.MAGIC, "Got bad getThumbnail:" + results[0]["image"])
        
    def testPathSearch(self):
        """ See that path search matches results correctly """
        
        self.loginOwner()
         
        self.portal.invokeFactory("Folder", id="level1")
        self.portal.level1.invokeFactory("Folder", id="level2")       
        self.portal.level1.level2.invokeFactory("Document", id="doc")       
        
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")                        
        
        fields = [
          # test query to return content beneath /level1 inc. /level1
          {
          "type" : "path",
          "name" : "Content path",
          "index" : "path",
          "default" : "/level1",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },          
        ]
                
        # create a custom search
        self.portal.sampleSearch.setFields(fields)           
        
        #print "performing search"
        
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()
                
        self.failUnless(len(results) == 3, "got results: %s" % str(results))
        
        fields = [
          # test query to return content beneath /level1/level2
          {
          "type" : "path",
          "name" : "Content path",
          "index" : "path",
          "default" : "level1/level2",
          "mode" : SearchField.MODE_FIXED_VALUE,
          },          
        ]
                
        # create a custom search
        self.portal.sampleSearch.setFields(fields)           

        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()
        
        self.failUnless(len(results) == 2, "got results: %s" % str(results))


    def testSorting(self):
        """ Test that custom sorting works propeply
        
        - Create test content
        - Check sorted search results 
        - Invert sort priorities, recheck results
        - Flip primary sort key to descending, recheck results
        """
    
        self.loginOwner()
      
        self.createTestCatalogsAndObjects()
        
        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")       
        
        # some basic "method is there" testing
        self.portal.sampleSearch.getSortKeyOrders()
        self.portal.sampleSearch.getSortKeyModes()        
                
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },                  
        ]
        self.portal.sampleSearch.setFields(fields) 
        
        sorting = [
           { 
          "name" : "amoutn",
          "index" : "getAmount",
          "priority" : "2",
          "mode" : SortKey.MODE_FORCED
          },                  
          {
          "name" : "price",
          "index" : "getPrice",
          "priority" : "1",
          "mode" : SortKey.MODE_FORCED
          },
        
        ]             
        self.portal.sampleSearch.setSorting(sorting)  
        
        skm = self.portal.sampleSearch.getSortKeyManager()        
        self.failUnless(len(skm.getSortKeys()) == 2)
        
        displayList = self.portal.sampleSearch.getUserSortingVocabulary()     
        self.failUnless(len(displayList) == 0)
  
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()        
        self.failUnless(len(results) >= 2)
        self.failUnless("/dummy1" in results[0]["url"], "results:" + str(results))
        self.failUnless("/dummy2" in results[1]["url"], "results:" + str(results))
                
        # Put sorting priorities to different order
        # this should inverse the result listing
        sorting = [
           { 
          "name" : "amount",
          "index" : "getAmount",
          "priority" : "1",
          "mode" : SortKey.MODE_FORCED
          },                  
          {
          "name" : "price",
          "index" : "getPrice",
          "priority" : "2",
          "mode" : SortKey.MODE_FORCED
          },        
        ]
        self.portal.sampleSearch.setSorting(sorting)  
        #spec = self.portal.sampleSearch.getSortKeyManager().constructSortSpecification(request=None, maxUserKeys=2)

        results = self.portal.sampleSearch.doCustomSearch()
        self.failUnless(len(results) >= 2)
        self.failUnless("/dummy2" in results[0]["url"], "results:" + str(results))
        self.failUnless("/dummy1" in results[1]["url"], "results:" + str(results))
        
        # then descending test
        sorting = [
           { 
          "name" : "amount",
          "index" : "getAmount",
          "priority" : "1",
          "mode" : SortKey.MODE_FORCED,
          "order" : SortKey.ORDER_DESCENDING,
          },                       
        ]
        self.portal.sampleSearch.setSorting(sorting)  
        results = self.portal.sampleSearch.doCustomSearch()
        self.failUnless(len(results) >= 2, "got results:" + str(results))
        self.failUnless("/dummy1" in results[0]["url"], "results:" + str(results))
        self.failUnless("/dummy2" in results[1]["url"], "results:" + str(results))        


    def xxtestIsTitleSortable(self):
        """ Test that sorting using Title works
        
        By default, AdvancedQuery cannot order ZCTextIndex
        indexes because they lack documentToMap attribute.
        
        """
    
        self.loginOwner()
      
        self.createTestCatalogsAndObjects()
        
        # create searched item
        self.portal.invokeFactory("CustomSearch", id="sampleSearch")       
        
        # some basic "method is there" testing
        self.portal.sampleSearch.getSortKeyOrders()
        self.portal.sampleSearch.getSortKeyModes()        
                
        fields = [
          { 
          "type" : "value",
          "name" : "Title",
          "index" : "Title",
          },                  
        ]
        self.portal.sampleSearch.setFields(fields) 
        
        sorting = [
           { 
          "name" : "title",
          "index" : "Title",
          "priority" : "2",
          "mode" : SortKey.MODE_FORCED
          },                  
        
        ]             
        self.portal.sampleSearch.setSorting(sorting)  
     
        # Now execute custom search
        results = self.portal.sampleSearch.doCustomSearch()        
        self.failUnless(len(results) >= 2)
        self.failUnless("/dummy1" in results[0]["url"], "results:" + str(results))
        self.failUnless("/dummy2" in results[1]["url"], "results:" + str(results))
                

  
    def loginOwner(self):
        try:
            self.loginPortalOwner() # Plone 2.1.x, fails on 2.5
        except:        
            self.loginAsPortalOwner() # Plone 2.5.x
        

    def createTestCatalogsAndObjects(self):
       
        # create a custom portal_catalog search index
        indexes = [
               ("getPrice", "FieldIndex", None),                      
               ("getAmount", "FieldIndex", None),                    
               ]
        
        metadata = ("getPrice",)
        
        runner = InstallationRunner(CatalogInstaller(indexes, metadata),)
        
        ic = InstallationContext(self.portal, globals())
  
        # Runs the installation ans return the report
        runner.install(ic, auto_reorder=1)    
        
        # Now create the test object
        mkDummyInContext(PriceDummyContent, oid='dummy1', context=self.portal,
                                       schema=ATImage.schema)

        mkDummyInContext(PriceDummyContent, oid='dummy2', context=self.portal,
                                       schema=ATImage.schema)
        
        self.portal.dummy1.setPrice(1000)
        self.portal.dummy2.setPrice(2000)
        self.portal.dummy1.setAmount(5)        
        self.portal.dummy2.setAmount(1)
        
        self.portal.dummy1.reindexObject()        
        self.portal.dummy2.reindexObject()     
        
    def isInResults(self, results, url):   
        """ Check if certain URL (part of it) is found in search result links """
        for r in results:
            if url in r["url"]:
                return True
        return False    
        
if __name__ == '__main__':
    framework()

# below some inactive debugging code

def getRolesInContext(self, object):
    """Return the list of roles assigned to the user,
       including local roles assigned in context of
       the passed in object."""
    userid=self.getId()
    roles=self.getRoles()
    local={}
    object=getattr(object, 'aq_inner', object)
    while 1:
        local_roles = getattr(object, '__ac_local_roles__', None)
        if local_roles:
            if callable(local_roles):
                local_roles=local_roles()
            dict=local_roles or {}
            for r in dict.get(userid, []):
                local[r]=1
        inner = getattr(object, 'aq_inner', object)
        parent = getattr(inner, 'aq_parent', None)
        if parent is not None:
            object = parent
            continue
        if hasattr(object, 'im_self'):
            object=object.im_self
            object=getattr(object, 'aq_inner', object)
            continue
        break
    roles=list(roles) + local.keys()
    return roles

def dumpPermissionsAndRoles(user, context):
    """ Print permission to role map to stdout
    
    List all permissions for the context object
    
    - Permission name
    
    - Is permissions acquired (inherit from the parent)
    
    - Locally set roles which have this permission    
    """
    
    list_type = type([])
    tuple_type = type([])
    
    all_permissions = context.ac_inherited_permissions(all=True)
    
    for p in all_permissions:
        name, value = p[:2]
        p=Permission(name,value,context)
        roles = p.getRoles()
        
        if type(roles) == list_type:
            print "Permission:" + name + " [Acquired] " + str(roles) 
        else:
            print "Permission:" + name + " [not acq] " + str(roles) 
